#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (C) 2010-2017  Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

import json
from jinja2 import TemplateSyntaxError

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.db.models import Q
from django.forms.utils import ErrorList
from django.http import HttpResponse, HttpResponseRedirect, Http404
from django.shortcuts import render, redirect
from django.urls import reverse
from django.views.generic import RedirectView
from ishtar_common.utils import ugettext_lazy as _, pgettext_lazy, parse_parcels
from django.forms import ValidationError

from archaeological_operations import models
from archaeological_operations import forms
from archaeological_operations.forms import SelectedParcelGeneralFormSet
from archaeological_operations import wizards
from ishtar_common.forms import ClosingDateFormSelection, FinalForm, FinalDeleteForm
from ishtar_common.models import (
    get_current_profile,
    IshtarSiteProfile,
    DocumentTemplate,
)
from archaeological_context_records.models import ContextRecord
from ishtar_common.utils import check_permissions_condition
from ishtar_common.views import (
    gen_generate_doc,
    QAItemEditForm,
    QABaseLockView,
    wizard_is_available,
    QAItemForm,
    IshtarMixin,
    LoginRequiredMixin,
)
from ishtar_common.views_item import get_item, show_item, revert_item, new_qa_item
from ishtar_common.wizards import SearchWizard


def autocomplete_patriarche(request):
    ishtaruser = getattr(request.user, "ishtaruser", None)
    if not ishtaruser:
        return HttpResponse(content_type="text/plain")
    if (
        not ishtaruser.has_permission("archaeological_operations.view_operation")
        and not ishtaruser.has_permission("archaeological_operations.view_own_operation")
    ):
        return HttpResponse(content_type="text/plain")
    if not request.GET.get("term"):
        return HttpResponse(content_type="text/plain")
    q = request.GET.get("term")
    query = Q()
    for q in q.split(" "):
        query &= Q(code_patriarche__startswith=q)
    limit = 15
    operations = models.Operation.objects.filter(query).order_by("code_patriarche")[
        :limit
    ]
    data = json.dumps(
        [
            {"id": operation.code_patriarche, "value": operation.code_patriarche}
            for operation in operations
        ]
    )
    return HttpResponse(data, content_type="text/plain")


def autocomplete_archaeologicalsite(request):
    ishtaruser = getattr(request.user, "ishtaruser", None)
    if not ishtaruser:
        return HttpResponse(content_type="text/plain")
    if not ishtaruser.has_permission(
        "archaeological_operations.view_archaeologicalsite"
    ) and not ishtaruser.has_permission(
        "archaeological_operations.view_own_archaeologicalsite",
    ):
        return HttpResponse(content_type="text/plain")
    if not request.GET.get("term"):
        return HttpResponse(content_type="text/plain")
    q = request.GET.get("term")
    query = Q()
    for q in q.split(" "):
        qt = Q(reference__icontains=q) | Q(name__icontains=q)
        query = query & qt
    limit = 15
    sites = (
        models.ArchaeologicalSite.objects.filter(query)
        .distinct()
        .order_by("reference")[:limit]
    )
    data = json.dumps([{"id": site.pk, "value": str(site)[:60]} for site in sites])
    return HttpResponse(data, content_type="text/plain")


new_archaeologicalsite = new_qa_item(
    models.ArchaeologicalSite,
    forms.ArchaeologicalSiteForm,
    many=True,
    page_name=_("New archaeological site"),
)


def autocomplete_operation(request):
    ishtaruser = getattr(request.user, "ishtaruser", None)
    if not ishtaruser:
        return HttpResponse(content_type="text/plain")
    # person_types = request.user.ishtaruser.person.person_type
    if (
        not ishtaruser.has_permission("archaeological_operations.view_operation")
        and not ishtaruser.has_permission(
            "archaeological_operations.view_own_operation"
        )
    ):
        return HttpResponse(content_type="text/plain")
    if not request.GET.get("term"):
        return HttpResponse(content_type="text/plain")
    q = request.GET.get("term")
    query = Q()
    for q in q.split(" "):
        extra = Q(towns__name__icontains=q) | Q(common_name__icontains=q)
        try:
            int(q)
            extra = extra | Q(year=q) | Q(operation_code=q)
        except ValueError:
            pass
        if settings.COUNTRY == "fr":
            if q.startswith("OA"):
                q = q[2:]
            try:
                int(q)
                extra |= Q(code_patriarche__contains=q)
            except ValueError:
                pass
        query = query & extra
    limit = 15
    operations = models.Operation.objects.filter(query).distinct()[:limit]
    data = json.dumps(
        [{"id": operation.pk, "value": str(operation)} for operation in operations]
    )
    return HttpResponse(data, content_type="text/plain")


def get_available_operation_code(request, year=None):
    ishtaruser = getattr(request.user, "ishtaruser", None)
    if not ishtaruser:
        return HttpResponse(content_type="text/plain")
    if not ishtaruser.has_permission(
            "archaeological_operations.view_operation"
    ) and not ishtaruser.has_permission(
        "archaeological_operations.view_own_operation"
    ):
        return HttpResponse(content_type="text/plain")
    data = json.dumps({"id": models.Operation.get_available_operation_code(year)})
    return HttpResponse(data, content_type="text/plain")


get_operation = get_item(
    models.Operation, "get_operation", "operation", search_form=forms.OperationSelect
)

show_operation = show_item(models.Operation, "operation")
revert_operation = revert_item(models.Operation)

get_administrativeactop = get_item(
    models.AdministrativeAct,
    "get_administrativeactop",
    "administrativeactop",
    base_request={"operation__pk__isnull": False},
)

get_administrativeact = get_item(
    models.AdministrativeAct, "get_administrativeact", "administrativeact"
)

show_administrativeact = show_item(models.AdministrativeAct, "administrativeact")


operation_search_wizard = wizards.OperationSearch.as_view(
    [("general-operation_search", forms.OperationFormSelection)],
    label=_("Operation search"),
    url_name="operation_search",
)


wizard_steps = [
    ("filechoice-operation_creation", forms.OperationFormFileChoice),
    ("general-operation_creation", forms.OperationFormGeneral),
    ("judiciary-operation_creation", forms.CourtOrderedSeizureForm),
    ("preventive-operation_creation", forms.OperationFormPreventive),
    ("preventivediag-operation_creation", forms.OperationFormPreventiveDiag),
    ("final-operation_creation", FinalForm),
]


def get_check_files_for_operation(other_check=None):
    def func(self):
        if not get_current_profile().files or \
                not check_permissions_condition(["archaeological_files.view_file"])(self):
            return False
        if not other_check:
            return True
        return other_check(self)

    return func


check_files_for_operation = get_check_files_for_operation()


ope_crea_condition_dict = {
    "filechoice-operation_creation": check_files_for_operation,
    "judiciary-operation_creation": wizards.is_judiciary(
        "general-operation_creation",
        models.OperationType,
        "operation_type",
    ),
    "preventive-operation_creation": get_check_files_for_operation(
        wizards.is_preventive(
            "general-operation_creation",
            models.OperationType,
            "operation_type",
            "prev_excavation",
        )
    ),
    "preventivediag-operation_creation": get_check_files_for_operation(
        wizards.is_preventive(
            "general-operation_creation",
            models.OperationType,
            "operation_type",
            "arch_diagnostic",
        )
    ),
    "parcelsgeneral-operation_creation": wizards.has_associated_file(
        "filechoice-operation_creation", negate=True
    ),  # TODO: AFAC
    "parcels-operation_creation": wizards.has_associated_file(
        "filechoice-operation_creation"
    ),  # TODO: AFAC
}

operation_creation_wizard = wizards.OperationWizard.as_view(
    wizard_steps,
    label=_("New operation"),
    condition_dict=ope_crea_condition_dict,
    url_name="operation_creation",
)

operation_modif_wizard_steps = [
    ("selec-operation_modification", forms.OperationFormSelection),
    ("general-operation_modification", forms.OperationFormModifGeneral),
    ("judiciary-operation_modification", forms.CourtOrderedSeizureForm),
    ("preventive-operation_modification", forms.OperationFormPreventive),
    ("preventivediag-operation_modification", forms.OperationFormPreventiveDiag),
    ("final-operation_modification", FinalForm),
]


ope_modif_condition_dict = {
    "preventive-operation_modification": get_check_files_for_operation(
        wizards.is_preventive(
            "general-operation_modification",
            models.OperationType,
            "operation_type",
            "prev_excavation",
        )
    ),
    "preventivediag-operation_modification": get_check_files_for_operation(
        wizards.is_preventive(
            "general-operation_modification",
            models.OperationType,
            "operation_type",
            "arch_diagnostic",
        )
    ),
    "judiciary-operation_modification": wizards.is_judiciary(
        "general-operation_modification",
        models.OperationType,
        "operation_type",
    ),
}

operation_modification_wizard = wizards.OperationModificationWizard.as_view(
    operation_modif_wizard_steps,
    label=_("Operation modification"),
    condition_dict=ope_modif_condition_dict,
    url_name="operation_modification",
)


def operation_modify(request, pk):
    if not wizard_is_available(
        operation_modification_wizard, request, models.Operation, pk
    ):
        return HttpResponseRedirect("/")

    wizard_url = "operation_modification"
    wizards.OperationModificationWizard.session_set_value(
        request, "selec-" + wizard_url, "pk", pk, reset=True
    )
    return redirect(reverse(wizard_url, kwargs={"step": "general-" + wizard_url}))


def operation_add(request, file_id):
    operation_creation_wizard(request)
    wizards.OperationWizard.session_set_value(
        request, "filechoice-operation_creation", "associated_file", file_id, reset=True
    )
    return redirect(
        reverse("operation_creation", kwargs={"step": "general-operation_creation"})
    )


operation_closing_steps = [
    ("selec-operation_closing", forms.OperationFormSelection),
    ("date-operation_closing", ClosingDateFormSelection),
    ("final-operation_closing", forms.FinalOperationClosingForm),
]

operation_closing_wizard = wizards.OperationClosingWizard.as_view(
    operation_closing_steps,
    label=_("Operation closing"),
    url_name="operation_closing",
)


operation_deletion_steps = [
    ("selec-operation_deletion", forms.OperationFormMultiSelection),
    ("final-operation_deletion", forms.OperationDeletionForm),
]

operation_deletion_wizard = wizards.OperationDeletionWizard.as_view(
    operation_deletion_steps,
    label=_("Operation deletion"),
    url_name="operation_deletion",
)


def operation_delete(request, pk):
    if not wizard_is_available(
        operation_deletion_wizard, request, models.Operation, pk
    ):
        return HttpResponseRedirect("/")

    wizard_url = "operation_deletion"
    wizards.OperationDeletionWizard.session_set_value(
        request, "selec-" + wizard_url, "pks", pk, reset=True
    )
    return redirect(reverse(wizard_url, kwargs={"step": "final-" + wizard_url}))


RELATION_FORMSET_EXTRA_FORM = 3


def get_parcel_modify(model, key_for_parcel, url_name):
    # TODO: simplify like get_relation_modify
    def _modify_parcels(request, pk):
        formset_class = SelectedParcelGeneralFormSet
        item = model.objects.get(pk=pk)
        parcels = models.Parcel.objects.filter(**{key_for_parcel: pk}).all()
        available_towns = [(town.pk, str(town)) for town in item.towns.all()]

        initial = []
        for parcel in parcels:
            town = models.Town.objects.get(pk=parcel.town_id)
            initial.append({
                "pk": parcel.pk,
                "town": town.pk,
                "year": parcel.year,
                "section": parcel.section,
                "parcel_number": parcel.parcel_number,
                "public_domain": parcel.public_domain
            })

        data = {
            'TOWNS': available_towns,
            'form-TOTAL_FORMS': len(initial) + RELATION_FORMSET_EXTRA_FORM,
            'form-INITIAL_FORMS': 0,
            'form-MIN_NUM_FORMS': 0,
            'form-MAX_NUM_FORMS': 100,
        }

        if request.method == 'POST':
            new_data = dict(request.POST)
            new_data = {k: new_data[k][0] for k in new_data}  # convert POST to classic dict
            new_data.update(data)
            formset = formset_class(new_data)
            parcel_selection = new_data.get('_parcel_selection', None)

            if parcel_selection:
                selected_town = new_data.get("_town")
                p_town = int(selected_town)
                _parcels = parse_parcels(parcel_selection)

                for p in _parcels:
                    p_year = p["year"]
                    p_section = p["section"]
                    p_parcel_number = p["parcel_number"]
                    query = {
                        "section": p_section,
                        "parcel_number": p_parcel_number,
                        "town_id": p_town,
                        key_for_parcel: item
                    }
                    q = models.Parcel.objects.filter(**query)
                    if not q.count():
                        added_parcel = models.Parcel.objects.create(
                            **{
                                "section": p_section,
                                "parcel_number": p_parcel_number,
                                "town_id": p_town,
                                key_for_parcel: item
                            }
                        )
                        if p_year:
                            added_parcel.year = p_year
                            added_parcel.save()
                        item.parcels.add(added_parcel)
                return redirect(reverse(url_name, args=[pk]))

            if formset.is_valid():
                is_valid = True
                for idx_form, data in enumerate(formset.cleaned_data):
                    if (not data.get('parcel_number') or not data.get('section')) and \
                            not data.get('public_domain'):
                        continue

                    if data.get("pk"):
                        try:
                            current_parcel = models.Parcel.objects.get(pk=data.get("pk"))
                        except models.Parcel.DoesNotExist:
                            continue
                        not_deleted_or_associated = True
                        for key, value in data.items():
                            if key == "DELETE" and value is True:
                                parcel_associated_by_context_record = \
                                    ContextRecord.objects.filter(parcel=current_parcel.pk).all()
                                if parcel_associated_by_context_record.count() > 0:
                                    is_valid = False
                                    formset.forms[idx_form].errors["__all__"] = ErrorList([
                                        ValidationError(
                                            _("This parcel is associated with a context "
                                              "record. It can't be deleted.")
                                        )
                                    ])
                                else:
                                    current_parcel.delete()
                                not_deleted_or_associated = False

                        if not_deleted_or_associated:
                            current_parcel.year = data.get("year")
                            current_parcel.section = data.get("section")
                            current_parcel.parcel_number = data.get("parcel_number")
                            current_parcel.town = models.Town.objects.get(pk=int(data.get("town")))
                            current_parcel.public_domain = data.get("public_domain")
                            current_parcel.save()
                    else:
                        parcel_to_add = models.Parcel.objects.create(
                            **{
                                "section": data.get("section"),
                                "parcel_number": data.get("parcel_number"),
                                "town_id": int(data.get("town")),
                                key_for_parcel: item,
                                "year": data.get("year"),
                                "public_domain":data.get("public_domain")
                            }
                        )
                        item.parcels.add(parcel_to_add)
                if is_valid:
                    return redirect(reverse(url_name, args=[pk]))
        else:
            formset = formset_class(initial=initial, data=data)

        return render(request, 'ishtar/forms/modify_parcels.html', {
            'formset': formset,
            "url": reverse(url_name, args=[pk])
        })
    return _modify_parcels


operation_modify_parcels = get_parcel_modify(
    models.Operation, "operation", "operation-parcels-modify"
)


def _formset_get_deleted(request, data, pk_key):
    new_data = dict(request.POST)
    new_data = {k: new_data[k][0] for k in new_data}  # convert POST to classic dict
    no_values = list(range(data["form-TOTAL_FORMS"]))
    deleted = {}
    for k, value in list(new_data.items()):
        if not value or not k.startswith("form-"):
            continue
        try:
            form_number = int(k.split("-")[1])
        except (ValueError, IndexError) as __:
            continue
        if k.endswith("-DELETE"):
            if new_data.get(f"form-{form_number}-{pk_key}", None):
                deleted[form_number] = new_data[f"form-{form_number}-{pk_key}"]
                if form_number not in no_values:  # put it back in no values
                    no_values.append(form_number)
            else:
                new_data.pop(k)
        elif form_number in no_values and form_number not in deleted:
            no_values.pop(no_values.index(form_number))
    for no_value in no_values:
        for k in list(new_data.keys()):
            if k.startswith(f"form-{no_value}-"):
                new_data.pop(k)
    data["form-TOTAL_FORMS"] = data["form-TOTAL_FORMS"] - len(no_values)
    new_data.update(data)
    return new_data, deleted


def get_relation_modify(model, model_relation, formset_class, url_name,
                        filter_operations=False):
    def _modify_relation(request, pk, current_right=None):
        try:
            item = model.objects.get(pk=pk)
        except model.DoesNotExist:
            raise Http404()
        if "_own_" in current_right:
            if not request.user.has_perm(current_right, item):
                raise PermissionDenied()
        elif current_right:
            if not request.user.has_perm(current_right):
                raise PermissionDenied()
        relations = model_relation.objects.filter(left_record_id=pk).all()
        form_kwargs = {"left_record": item}

        items, current_items = [], []
        if filter_operations:
            items = [
                (item.id, str(item))
                for item in model.objects.filter(operation=item.operation).all()
            ]
            current_items = [item[0] for item in items]

        initial = []
        for relation in relations:
            initial.append({
                "pk": relation.pk,
                "right_record": relation.right_record_id,
                "relation_type": relation.relation_type_id,
            })
            if filter_operations and relation.right_record_id not in current_items:
                items.append((relation.right_record_id, str(relation.right_record)))
        data = {
            'form-TOTAL_FORMS': len(initial) + RELATION_FORMSET_EXTRA_FORM,
            'form-INITIAL_FORMS': 0,
            'form-MIN_NUM_FORMS': 0,
            'form-MAX_NUM_FORMS': 100,
        }
        if filter_operations:
            data["CURRENT_ITEMS"] = items
        if request.method == 'POST':
            # remove empty lines and get deleted
            new_data, deleted = _formset_get_deleted(request, data, "pk")
            formset = formset_class(data=new_data, form_kwargs=form_kwargs)
            if formset.is_valid():
                is_valid = True
                # delete
                for deleted_id in deleted.values():
                    try:
                        model_relation.objects.get(pk=deleted_id).delete()
                    except model_relation.DoesNotExist:
                        continue

                for idx_form, data in enumerate(formset.cleaned_data):
                    if not data.get('right_record') or not data.get('relation_type'):
                        continue

                    if data.get("pk"):
                        try:
                            current_relation = model_relation.objects.get(pk=data.get("pk"))
                        except model_relation.DoesNotExist:
                            continue
                        not_deleted_or_associated = True
                        for key, value in data.items():
                            if key == "DELETE" and value is True:
                                current_relation.delete()
                                not_deleted_or_associated = False

                        if not_deleted_or_associated:
                            current_relation.right_record_id = data.get("right_record")
                            current_relation.relation_type_id = data.get("relation_type")
                            current_relation.save()
                    else:
                        model_relation.objects.create(
                            **{
                                "left_record_id": item.pk,
                                "right_record_id": data.get("right_record"),
                                "relation_type_id": data.get("relation_type"),
                            }
                        )
                if is_valid:
                    return redirect(reverse(url_name, args=[pk]))
        else:
            formset = formset_class(initial=initial, data=data, form_kwargs=form_kwargs)

        return render(request, 'ishtar/forms/modify_relations.html', {
            'formset': formset,
            "url": reverse(url_name, args=[pk])
        })
    return _modify_relation


operation_modify_relations = get_relation_modify(
    models.Operation, models.RecordRelations,
    forms.RecordRelationsFormSet, "operation-relation-modify"
)


RELATION_LIMIT = 50


def operation_site_modify(model, related_model, related_key, formset_class, url_name):

    def _get_initial(q_relations):
        initial = []
        current_relation_ids = q_relations.values_list("id", flat=True)
        excess = current_relation_ids.count() > RELATION_LIMIT and RELATION_LIMIT
        for relation_id in current_relation_ids[:RELATION_LIMIT]:
            initial.append({
                "right_record": relation_id,
            })
        return initial, current_relation_ids, excess
    get_initial = _get_initial

    def view(request, pk, window_id=None, current_right=None):
        try:
            item = model.objects.get(pk=pk)
        except model.DoesNotExist:
            raise Http404()
        if "_own_" in current_right:
            if not item.is_own(request.user):
                raise PermissionDenied()
        q_relations = getattr(item, related_key)
        initial, current_relation_ids, excess = get_initial(q_relations)
        if request.method == 'POST':
            formset = formset_class(data=request.POST)
            if formset.is_valid():
                deleted_ids = []
                for form in formset.deleted_forms:
                    deleted_id = form.cleaned_data.get("right_record", None)
                    try:
                        deleted_id = int(deleted_id)
                    except (ValueError, TypeError):
                        continue
                    deleted_ids.append(deleted_id)
                    if deleted_id in current_relation_ids:
                        try:
                            del_item = related_model.objects.get(pk=deleted_id)
                        except related_model.DoesNotExist:
                            continue
                        q_relations.remove(del_item)

                for idx_form, data in enumerate(formset.cleaned_data):
                    if not data.get('right_record'):
                        continue
                    try:
                        new_item_id = int(data.get("right_record"))
                        if new_item_id in deleted_ids or new_item_id in current_relation_ids:
                            continue
                        new_item = related_model.objects.get(pk=new_item_id)
                    except (ValueError, TypeError, related_model.DoesNotExist):
                        continue
                    q_relations.add(new_item)
                initial, __, __ = get_initial(q_relations)
                formset = formset_class(initial=initial)
                item.update_search_vector()
        else:
            formset = formset_class(initial=initial)

        return render(request, 'ishtar/forms/modify_relations.html', {
            'formset': formset,
            "url": reverse(url_name, args=[pk]),
            "window_id": window_id,
            "excess": str(
                _("Too many relations. Only the first {limit} items are displayed.")
            ).format(limit=excess) if excess else None
        })
    return view


operation_site_modify_relations = operation_site_modify(models.Operation, models.ArchaeologicalSite, "archaeological_sites", forms.OpeSiteRelationsFormSet, "operation-site-relations-modify")
site_operation_modify_relations = operation_site_modify(models.ArchaeologicalSite, models.Operation, "operations", forms.SiteOpeRelationsFormSet, "site-operation-relations-modify")


# archaeological sites


def site_extra_context(request, item):
    return {"SITE_LABEL": IshtarSiteProfile.get_default_site_label()}


get_site = get_item(
    models.ArchaeologicalSite, "get_site", "site", search_form=forms.SiteSelect
)
show_site = show_item(models.ArchaeologicalSite, "site", extra_dct=site_extra_context)
revert_site = revert_item(models.ArchaeologicalSite)

site_search_wizard = wizards.SiteSearch.as_view(
    [("general-site_search", forms.SiteFormSelection)],
    url_name="site_search",
)


site_creation_steps = [
    ("general-site_creation", forms.SiteForm),
    ("underwater-site_creation", forms.SiteUnderwaterForm),
    ("final-site_creation", FinalForm),
]

site_creation_wizard = wizards.SiteWizard.as_view(
    site_creation_steps,
    condition_dict={"underwater-site_creation": forms.check_underwater_module},
    url_name="site_creation",
)

site_modification_steps = [
    ("selec-site_modification", forms.SiteFormSelection),
    ("general-site_modification", forms.SiteForm),
    ("underwater-site_modification", forms.SiteUnderwaterForm),
    ("final-site_modification", FinalForm),
]

site_modification_wizard = wizards.SiteModificationWizard.as_view(
    site_modification_steps,
    condition_dict={"underwater-site_modification": forms.check_underwater_module},
    url_name="site_modification",
)


def site_modify(request, pk):
    if not wizard_is_available(
        site_modification_wizard, request, models.ArchaeologicalSite, pk
    ):
        return HttpResponseRedirect("/")
    wizard_url = "site_modification"
    wizards.SiteModificationWizard.session_set_value(
        request, "selec-" + wizard_url, "pk", pk, reset=True
    )
    return redirect(reverse(wizard_url, kwargs={"step": "general-" + wizard_url}))


site_deletion_steps = [
    ("selec-site_deletion", forms.SiteFormMultiSelection),
    ("final-site_deletion", FinalDeleteForm),
]

site_deletion_wizard = wizards.SiteDeletionWizard.as_view(
    site_deletion_steps,
    label=_("Site deletion"),
    url_name="site_deletion",
)


def site_delete(request, pk):
    if not wizard_is_available(
        site_deletion_wizard, request, models.ArchaeologicalSite, pk
    ):
        return HttpResponseRedirect("/")
    wizard_url = "site_deletion"
    wizards.SiteDeletionWizard.session_set_value(
        request, "selec-" + wizard_url, "pks", pk, reset=True
    )
    return redirect(reverse(wizard_url, kwargs={"step": "final-" + wizard_url}))


def site_add_operation(request, pks, current_right=None):
    pks = [int(p) for p in pks.split("-")]
    for pk in pks:
        q = models.ArchaeologicalSite.objects.filter(pk=pk)
        if not q.count():
            raise Http404()
        site = q.all()[0]
        if not site.can_do(request, "view", app="archaeological_operations",
                           model_name="archaeologicalsite") \
                or site.operations.count():
            raise Http404()
    # operation add permission checked on view call
    operation_creation_wizard(request)
    wizards.OperationWizard.session_set_value(
            request, "general-operation_creation", "archaeological_site", pks,
            reset=True
    )
    return redirect(
        reverse("operation_creation", kwargs={"step": "general-operation_creation"})
    )


def site_add_top_operation(request, pks, current_right=None):
    q = models.ArchaeologicalSite.objects.filter(pk=pks)
    if not q.count():
        raise Http404()
    site = q.all()[0]
    if not site.can_do(request, "view", app="archaeological_operations",
                       model_name="archaeologicalsite") \
            or not site.operations.count():
        raise Http404()
    # operation add permission checked on view call
    top_operation = site.create_or_update_top_operation(create=True)
    if not wizard_is_available(
            operation_modification_wizard, request, models.Operation, top_operation.pk
    ):
        return HttpResponseRedirect("/")

    wizard_url = "operation_modification"
    wizards.OperationModificationWizard.session_set_value(
        request, "selec-" + wizard_url, "pk", top_operation.pk, reset=True
    )
    return redirect(reverse(wizard_url, kwargs={"step": "general-" + wizard_url}))


operation_administrativeactop_search_wizard = wizards.SearchWizard.as_view(
    [
        (
            "general-operation_administrativeactop_search",
            forms.AdministrativeActOpeFormSelection,
        )
    ],
    label=_("Administrative act search"),
    url_name="operation_administrativeactop_search",
)

administrativeactop_steps = [
    ("selec-operation_administrativeactop", forms.OperationFormSelection),
    ("administrativeact-operation_administrativeactop", forms.AdministrativeActOpeForm),
    ("final-operation_administrativeactop", FinalForm),
]


operation_administrativeactop_wizard = wizards.OperationAdministrativeActWizard.as_view(
    administrativeactop_steps,
    label=_("Operation: new administrative act"),
    url_name="operation_administrativeactop",
)

operation_administrativeactop_modification_wizard = (
    wizards.OperationEditAdministrativeActWizard.as_view(
        [
            (
                "selec-operation_administrativeactop_modification",
                forms.AdministrativeActOpeFormSelection,
            ),
            (
                "administrativeact-operation_administrativeactop_modification",
                forms.AdministrativeActOpeModifForm,
            ),
            ("final-operation_administrativeactop_modification", FinalForm),
        ],
        label=_("Operation: administrative act modification"),
        url_name="operation_administrativeactop_modification",
    )
)


def operation_administrativeactop_modify(request, pk):
    if not wizard_is_available(
        operation_administrativeactop_modification_wizard,
        request,
        models.AdministrativeAct,
        pk,
    ):
        return HttpResponseRedirect("/")
    wizard_url = "operation_administrativeactop_modification"
    wizards.OperationEditAdministrativeActWizard.session_set_value(
        request, "selec-" + wizard_url, "pk", pk, reset=True
    )
    return redirect(
        reverse(wizard_url, kwargs={"step": "administrativeact-" + wizard_url})
    )


operation_administrativeactop_deletion_wizard = (
    wizards.AdministrativeActDeletionWizard.as_view(
        [
            (
                "selec-operation_administrativeactop_deletion",
                forms.AdministrativeActOpeFormSelection,
            ),
            (
                "final-operation_administrativeactop_deletion",
                forms.FinalAdministrativeActDeleteForm,
            ),
        ],
        label=_("Operation: administrative act deletion"),
        url_name="operation_administrativeactop_deletion",
    )
)


def operation_administrativeactop_delete(request, pk):
    if not wizard_is_available(
        operation_administrativeactop_deletion_wizard,
        request,
        models.AdministrativeAct,
        pk,
    ):
        return HttpResponseRedirect("/")
    wizard_url = "operation_administrativeactop_deletion"
    wizards.AdministrativeActDeletionWizard.session_set_value(
        request, "selec-" + wizard_url, "pk", pk, reset=True
    )
    return redirect(reverse(wizard_url, kwargs={"step": "final-" + wizard_url}))


def operation_adminact_add(request, pk, current_right=None):
    try:
        models.Operation.objects.get(pk=pk)
    except models.Operation.DoesNotExist:
        raise Http404()
    operation_administrativeactop_wizard(request)

    wizards.OperationAdministrativeActWizard.session_set_value(
        request, "selec-operation_administrativeactop", "pk", pk, reset=True
    )
    return redirect(
        reverse(
            "operation_administrativeactop",
            kwargs={"step": "administrativeact-operation_administrativeactop"},
        )
    )


administrativact_register_wizard = SearchWizard.as_view(
    [
        (
            "general-administrativact_register",
            forms.AdministrativeActRegisterFormSelection,
        )
    ],
    label=pgettext_lazy("admin act register", "Register"),
    url_name="administrativact_register",
)


generatedoc_administrativeactop = gen_generate_doc(models.AdministrativeAct)


def administrativeactfile_document(
    request, file=False, treatment=False, treatment_file=False
):
    search_form = forms.AdministrativeActOpeFormSelection
    document_type = "O"
    if file:
        from archaeological_files.forms import AdministrativeActFileFormSelection

        search_form = AdministrativeActFileFormSelection
        document_type = "F"
    elif treatment:
        from archaeological_finds.forms import AdministrativeActTreatmentFormSelection

        search_form = AdministrativeActTreatmentFormSelection
        document_type = "T"
    elif treatment_file:
        from archaeological_finds.forms import (
            AdministrativeActTreatmentFileFormSelection,
        )

        search_form = AdministrativeActTreatmentFileFormSelection
        document_type = "TF"

    ishtaruser = getattr(request.user, "ishtaruser", None)
    if not ishtaruser:
        return HttpResponse(content_type="text/plain")
    if not ishtaruser.has_permission("archaeological_operations.view_administrativeact"):
        return HttpResponse(content_type="text/plain")
    dct = {}
    DocumentGenerationAdminActForm = forms.DocumentGenerationAdminActForm
    if request.POST:
        dct["search_form"] = search_form(request.POST)
        dct["template_form"] = DocumentGenerationAdminActForm(
            document_type=document_type
        )
        c_object = None
        try:
            if dct["search_form"].is_valid():
                c_object = DocumentGenerationAdminActForm._associated_model.objects.get(
                    pk=dct["search_form"].cleaned_data.get("pk")
                )
        except DocumentGenerationAdminActForm._associated_model.DoesNotExist:
            pass
        if c_object:
            dct["template_form"] = DocumentGenerationAdminActForm(
                request.POST, document_type=document_type, obj=c_object
            )
            if dct["template_form"].is_valid():
                try:
                    return generatedoc_administrativeactop(
                        request,
                        dct["search_form"].cleaned_data.get("pk"),
                        dct["template_form"].cleaned_data.get("document_template"),
                    )
                except TemplateSyntaxError:
                    dct["search_form"] = search_form()
                    try:
                        template = DocumentTemplate.objects.get(
                            pk=dct["template_form"].cleaned_data.get(
                                "document_template"
                            )
                        ).name
                    except DocumentTemplate.DoesNotExist:
                        template = ""
                    dct["template_form"] = DocumentGenerationAdminActForm(
                        document_type=document_type
                    )
                    dct["template_error"] = str(
                        _(
                            'Syntax error on the source template "{}" - '
                            "contact your administrator and ask him to check "
                            "the syntax of this document."
                        )
                    ).format(template)
    else:
        dct["search_form"] = search_form()
        dct["template_form"] = DocumentGenerationAdminActForm(
            document_type=document_type
        )
    return render(request, "ishtar/administrativeact_document.html", dct)


def autocomplete_administrativeact(request):
    ishtaruser = getattr(request.user, "ishtaruser", None)
    if not ishtaruser:
        return HttpResponse(content_type="text/plain")
    if not ishtaruser.has_permission(
        "archaeological_operations.view_administrativeact"
    ) and not ishtaruser.has_permission(
        "archaeological_operations.view_own_administrativeact"
    ):
        return HttpResponse(content_type="text/plain")
    if not request.GET.get("term"):
        return HttpResponse(content_type="text/plain")
    q = request.GET.get("term")
    query = Q()
    for q in q.split(" "):
        qt = Q(act_type__label__icontains=q) | Q(towns_label=q)
        try:
            if len(q) == 4:
                qt |= Q(year=int(q))
            qt |= Q(index=int(q))
        except ValueError:
            pass
        query = query & qt
    limit = 15
    items = (
        models.AdministrativeAct.objects.filter(query)
        .order_by("year", "index")
        .distinct()[:limit]
    )
    data = json.dumps(
        [{"id": item.pk, "value": str(item)[:80] + " (...)"} for item in items]
    )
    return HttpResponse(data, content_type="text/plain")


def reset_wizards(request):
    for wizard_class, url_name in (
        (wizards.OperationWizard, "operation_creation"),
        (wizards.OperationModificationWizard, "operation_modification"),
        (wizards.OperationClosingWizard, "operation_closing"),
        (wizards.OperationDeletionWizard, "operation_deletion_wizard"),
        (wizards.OperationAdministrativeActWizard, "operation_administrativeactop"),
        (
            wizards.OperationEditAdministrativeActWizard,
            "operation_administrativeactop_modification",
        ),
        (
            wizards.AdministrativeActDeletionWizard,
            "operation_administrativeactop_deletion",
        ),
    ):
        wizard_class.session_reset(request, url_name)


class QAOperationForm(QAItemEditForm):
    model = models.Operation
    form_class = forms.QAOperationFormMulti


class QAOperationLockView(QABaseLockView):
    model = models.Operation
    base_url = "operation-qa-lock"


class QASiteLockView(QABaseLockView):
    model = models.ArchaeologicalSite
    base_url = "site-qa-lock"


class QAOperationdDuplicateFormView(QAItemForm):
    template_name = "ishtar/forms/qa_operation_duplicate.html"
    model = models.Operation
    page_name = _("Duplicate")
    form_class = forms.QAOperationDuplicateForm
    base_url = "operation-qa-duplicate"

    def get_form_kwargs(self):
        kwargs = super(QAOperationdDuplicateFormView, self).get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def form_valid(self, form):
        form.save()
        return HttpResponseRedirect(reverse("success"))

    def get_context_data(self, **kwargs):
        data = super(QAOperationdDuplicateFormView, self).get_context_data(**kwargs)
        data["action_name"] = _("Duplicate")
        return data


class QAArchaeologicalSiteDuplicateFormView(QAItemForm):
    template_name = "ishtar/forms/qa_site_duplicate.html"
    model = models.ArchaeologicalSite
    page_name = _("Duplicate")
    form_class = forms.QAArchaeologicalSiteDuplicateForm
    base_url = "site-qa-duplicate"

    def get_form_kwargs(self):
        kwargs = super(QAArchaeologicalSiteDuplicateFormView, self).get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def form_valid(self, form):
        form.save()
        return HttpResponseRedirect(reverse("success"))

    def get_context_data(self, **kwargs):
        data = super(QAArchaeologicalSiteDuplicateFormView, self).get_context_data(
            **kwargs
        )
        data["action_name"] = _("Duplicate")
        return data


class QAArchaeologicalSiteForm(QAItemEditForm):
    model = models.ArchaeologicalSite
    form_class = forms.QAArchaeologicalSiteFormMulti


class GenerateStatsOperation(IshtarMixin, LoginRequiredMixin, RedirectView):
    model = models.Operation

    def get_redirect_url(self, *args, **kwargs):
        return (
            reverse("display-item", args=[self.model.SLUG, self.item.pk])
            + "#statistics"
        )

    def get(self, request, *args, **kwargs):
        self.item = self.model.objects.get(pk=kwargs["pk"])
        self.item._get_or_set_stats("_nb_acts", update=True)
        self.item._get_or_set_stats("_nb_indexed_acts", update=True)
        self.item._get_or_set_stats("_nb_context_records", update=True)
        self.item._get_or_set_stats(
            "_nb_context_records_by_type", update=True, expected_type=list
        )
        self.item._get_or_set_stats(
            "_nb_context_records_by_periods", update=True, expected_type=list
        )
        self.item._get_or_set_stats("_nb_finds", update=True)
        self.item._get_or_set_stats(
            "_nb_finds_by_material_type", update=True, expected_type=list
        )
        self.item._get_or_set_stats(
            "_nb_finds_by_types", update=True, expected_type=list
        )
        self.item._get_or_set_stats(
            "_nb_finds_by_periods", update=True, expected_type=list
        )
        self.item._get_or_set_stats("_nb_documents", update=True)
        self.item._get_or_set_stats(
            "_nb_documents_by_types", update=True, expected_type=list
        )
        self.item._get_or_set_stats("_nb_stats_finds_by_ue", update=True)

        return super(GenerateStatsOperation, self).get(request, *args, **kwargs)
